import sys, os
import urllib
import socket
import re

import requests

from bldeif.connection import BLDConnection
from bldeif.utils.eif_exception import ConfigurationError, OperationalError
from bldeif.utils.chronuti import TimeStamp, ISO_FRAC_FORMAT, ISO_FORMAT
from bldeif.utils.status_matchmaker import Matchmaker

quote = urllib.parse.quote

############################################################################################

__version__ = "1.1.3"

############################################################################################

class BambooConnection(BLDConnection):
    def __init__(self, config, logger):
        super().__init__(logger)
        self.bamboo = None
        self.internalizeConfig(config)
        self.backend_version = ""
        self.username_required = False
        self.password_required = False
        self.plans = [] #former self.inventory

    def name(self):
        return "Bamboo"

    def version(self):
        global __version__
        return __version__

    def getBackendVersion(self):
        """
            Conform to Connection subclass protocol and provide the version of the system
            we are "connecting" to.
        """
        return self.backend_version

    def internalizeConfig(self, config):
        super().internalizeConfig(config)
        self.protocol   = config.get('Protocol', 'http')
        self.server     = config.get('Server', socket.gethostname())
        self.port       = config.get('Port', 8080)
        self.prefix     = config.get("Prefix", '')
        self.username   = config.get("Username", '')
        self.password   = config.get("Password", '')
        self.debug      = config.get("Debug", False)
        self.max_items  = config.get("MaxItems", 1000)
        self.projects   = []
        self.ac_project = config.get("AgileCentral_DefaultBuildProject", None)

        self.base_url   = "{0}://{1}:{2}/rest/api/latest".format(self.protocol, self.server, self.port)
        self.creds = (self.username, self.password)

        self.http_proxy = {}
        if self.proxy_server:
            proxy  = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
            if self.proxy_username and self.proxy_password:
                proxy  = "%s://%s:%s@%s:%s" % (self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)
            self.http_proxy = {self.protocol : proxy}
            self.log.info("Proxy for Bamboo connection:  %s" % proxy)

        valid_config_items = ['Server', 'Protocol', 'Prefix', 'Port',
                              'Username', 'User', 'Password',
                              'ProxyProtocol', 'ProxyServer', 'ProxyPort', 'ProxyUser', 'ProxyUsername',
                              'ProxyPassword',
                              'Debug', 'Lookback',
                              'AgileCentral_DefaultBuildProject',
                              'Projects'
                             ]

        invalid_config_items = [item for item in config.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "Bamboo section of the config contained these invalid entries: %s" % ", ".join(
                invalid_config_items)
            raise ConfigurationError(problem)

        if not config.get('Projects', False):
            problem = "No Projects sub-section was provided in your configuration in the Bamboo section"
            raise ConfigurationError(problem)

        for proj_section in config.get("Projects"):
            if not proj_section.get('Project', False):
                problem = "Subsection under Projects fails to list a Bamboo Project key or value, %s" % (repr(proj_section))
                raise ConfigurationError(problem)
            project = proj_section['Project']
            p = {}
            details = {}
            details['Plans'] = proj_section.get('Plans', [])
            if 'AgileCentral_Project' in proj_section:
                details['AgileCentral_Project'] = proj_section['AgileCentral_Project']
            else:
                details['AgileCentral_Project'] = self.ac_project
            p[project] = details
            self.projects.append(p)

        self.builds = {}


    def connect(self):
        self.log.info("Connecting to Bamboo")
        self.backend_version = self._getBambooVersion()
        self.log.info("Connected to Bamboo server: %s running at version %s" % (self.server, self.backend_version))
        self.log.info("Url: %s" % self.base_url)
        return True

    def _getBambooVersion(self):
        version  = None
        response = None
        msg = ''

        bamboo_url = "%s/info.json" % self.base_url
        headers = {'Content-Type': 'application/json'}
        self.log.debug(bamboo_url)
        try:
            response = requests.get(bamboo_url, auth=self.creds, proxies=self.http_proxy, headers=headers)
        except Exception as msg:
            self.log.error(msg)
        if response.status_code >= 300:
            raise ConfigurationError('%s  status_code: %s' % (msg, response.status_code))

        # self.log.debug(response.headers)
        result = response.json()
        if 'version' in result:
            return result['version']

    def disconnect(self):
        self.bamboo = None

    def validate(self):
        """
            Make sure any requisite conditions are satisfied.
            Are credentials needed and supplied?
        """
        satisfactory = True
        if self.username:
            self.log.debug(
                '%s - user entry "%s" detected in config file' % (self.__class__.__name__, self.username))
        else:
            self.log.error("No Username was provided in your configuration in the Bamboo section")
            return False

        if self.password:
            self.log.debug('%s - password entry detected in config file' % self.__class__.__name__)
        else:
            self.log.error("No Password was provided in your configuration in the Bamboo section")
            return False

        if not self.projects:
            self.log.error("No Projects were provided in your configuration in the Bamboo section")
            return False

        return satisfactory

    def targetActivityUnit(self):
        return 'plan'

    def getRecentBuilds(self, ref_ts, get_vcs_data=False):
        self.log.info("recent Builds criteria: CompletionTime >= %s" % (ref_ts.asLogStamp()))
        all_projects = self.getProjects()

        self.getPlans(all_projects)

        for plan in self.plans:
            self.getBuildsPerPlan(plan.key, ref_ts, get_vcs_data)
        total_builds_count = 0
        for ac_project, plan_builds in self.builds.items():
            for plan, builds in plan_builds.items():
                total_builds_count += len(builds)
        self.log.info("%d recently added %s builds detected" % (total_builds_count, self.name()))
        return self.builds


    def getProjects(self):
        """
            Use Bamboo REST API endpoint to obtain all visible/accessible Projects and their Plans:
                curl --user toto:totogithub http://localhost:8085/rest/api/latest/project.json?expand=projects.project.plans | python -m json.tool
        """
        all_projects = []
        endpoint = 'project.json?expand=projects.project.plans'
        headers = {'Content-Type': 'application/json'}
        url = "%s/%s" % (self.base_url, endpoint)
        response = requests.get(url, auth=self.creds, proxies=self.http_proxy, headers=headers)
        if response.status_code == 200:
            result = response.json()
            all_projects = result['projects']['project']
        return all_projects

    def findQualifiers(self, plan_name, entry):
        entry = entry.split(",")

        for exp in entry:
            pattern = re.compile('(%s)' % exp.replace('*', '.*'))
            m = pattern.search(plan_name)
            if m:
                return [m.string]

        return []

    def findMatchingProjectConfigData(self, project_name):
        config_project_data = [proj[project_name] for proj in self.projects if list(proj.keys())[0] == project_name][0]
        return config_project_data['Plans'] if 'Plans' in config_project_data else None

    def getPlans(self, all_projects):
        '''
            all_projects is a json of response from endpoint: 'project.json?expand=projects.project.plans'
            selected_projects is a subset of all_projects
            self.projects is a list of dictionaries of config entires under "Projects" property
        '''
        specified_plans = []
        excluded_plans  = []

        conf_project_names = [proj_name for project in self.projects for proj_name in project.keys()]
        selected_projects  = [project   for project in all_projects if project['name'] in conf_project_names]

        for project in selected_projects:
            bamboo_project_name = project['name']
            project_plans       = project['plans']['plan']
            if not project_plans:
                continue
            config_project_data = self.findMatchingProjectConfigData(bamboo_project_name)
            if not config_project_data:  # no Plans and subitems present in config for this project, means grab all Bamboo Plans for this Project
                project_plan_instances = [BambooPlan(raw_plan) for raw_plan in project_plans]
                self.plans.extend(project_plan_instances)
            else:
                plan_items     = [item['Plan']    for item in config_project_data if list(item.keys())[0] == 'Plan'   ]  # we could get one or several of these or none
                included_items = [item['include'] for item in config_project_data if list(item.keys())[0] == 'include']  # we could get one or several of these or none
                excluded_items = [item['exclude'] for item in config_project_data if list(item.keys())[0] == 'exclude']  # we could get one or several of these or none

                specified_plans = plan_items[:]

                if included_items:
                    inclusions = []
                    for inclusion_spec in included_items:
                        for raw_plan in project_plans:
                            inclusion = self.findQualifiers(raw_plan['shortName'], inclusion_spec)
                            inclusions.extend(inclusion)
                        if inclusions:
                            specified_plans.extend(inclusions)

                if excluded_items:
                    excluded_plans = []
                    for exclusion_spec in excluded_items:
                        for raw_plan in project_plans:
                            exclusion = self.findQualifiers(raw_plan['shortName'], exclusion_spec)
                            excluded_plans.extend(exclusion)


                qualifying_plans = list(set(specified_plans) - set(excluded_plans))
                for raw_plan in project_plans:
                    if raw_plan['shortName'] in qualifying_plans:
                        self.plans.append(BambooPlan(raw_plan))

        return self.plans


    def showQualifiedPlans(self):
        """
            Roll through self.plans to show all the plans
        """
        [print(str(plan)) for plan in self.plans]

    def getBuildsPerPlan(self, key, ref_ts, get_vcs_data):
        """
        curl --user toto:totogithub -g http://localhost:8085/rest/api/latest/result/FER-DON.json?expand=results[0:5].result | python -m json.tool

        to get more details on vcsRevisions:

        curl --user toto:totogithub -g http://localhost:8085/rest/api/latest/result/FER-DON.json?expand=results[0:5].result.vcsRevisions | python -m json.tool

        vcs data seeps to be limited to:
            "vcsRevision": [
                        {
                            "repositoryId": 360450,
                            "repositoryName": "bamboo-camillo",
                            "vcsRevisionKey": "561b474ef508710944574f4e33ea9f77a2abf69b"
                        }
                    ]
         Is it possible to get revision's commit message and timestamp?
         Yes, but only by hitting a separate endpoint per build result, e.g.
         http://localhost:8085/rest/api/latest/result/FER-DON/74.json?expand=changes.change.files,vcsRevisions


         Paging:

         endpoint = 'result/%s.json?expand=results[0:200].result.vcsRevisions' % key
         results[0:max] returns 25 results (one page) per plan. If there are more than 25 results per plan paging
         will be necessary. E.g. if one plan has 70 results, another 80, third 10, only 60 (25+25+10) will be returned
         instead of paging, use max-results=1000000 with arbitrary large number
         E.g. the endpoint below will return all 160 (80+70+10) results
         endpoint = 'result/%s.json?expand=results.result.vcsRevisions&max-results=1000000' % key
        """
        raw_builds = []
        endpoint = 'result/%s.json?expand=results.result.vcsRevisions&max-results=1000000' % key
        headers = {'Content-Type': 'application/json'}
        url = "%s/%s" % (self.base_url, endpoint)
        response = requests.get(url, auth=self.creds, proxies=self.http_proxy, headers=headers)
        if response.status_code == 200:
            result = response.json()
            raw_builds = result['results']['result']
            if raw_builds:
                self.extractQualifyingBuilds(raw_builds, ref_ts, get_vcs_data)

    def extractQualifyingBuilds(self, raw_builds, ref_ts, get_vcs_data):
        build_count = 0
        ac_project  = ''
        plan        = ''

        for record in raw_builds:
            build_timestamp = TimeStamp.fromBambooTimestampString(record['buildCompletedTime'], ISO_FRAC_FORMAT)
            if build_timestamp.asEpochSeconds() >= ref_ts.asEpochSeconds():
                build_count += 1
                # prep builds dict when there is at least one qualified build:
                if build_count == 1:
                    ac_project = self.getAgileCentralProject(record['projectName'])
                    if ac_project not in self.builds:
                        self.builds[ac_project] = {}
                    plan = BambooPlan(record['plan'])
                    self.builds[ac_project][plan] = []

                build = BambooBuild(record)
                build.changesets = self.getChangesetInfo(build) if get_vcs_data else []
                self.builds[ac_project][plan].append(build)
        if build_count > 1:
            self.builds[ac_project][plan] = self.builds[ac_project][plan][::-1]


    def getAgileCentralProject(self, bamboo_project_name):
        ac_projects = [ac_proj for project in self.projects
                               for key, proj_data in project.items() if key == bamboo_project_name
                               for key, ac_proj in proj_data.items() if key == 'AgileCentral_Project']
        if ac_projects:
            return ac_projects[0]

    def dumpTargets(self):
        for plan in self.plans:
            self.log.debug('Plan: %s' % plan)


    def _getTargetProjects(self, bld_conf, target_project_key, default_project):
        target_projects = [default_project]
        for proj_section in bld_conf["Projects"]:
            project = default_project
            if target_project_key in proj_section:
                project = proj_section[target_project_key]
            target_projects.append(project)
        return list(set(target_projects))


    def _identifyUnrecordedBuilds(self, recorded_builds, bld_builds):
        """
            If there are items in the recorded_builds for which there is  a counterpart in
            the bld_builds, the information has already been reflected in Agile Central.  --> NOOP

            If there are items in the bld_builds   for which there is no counterpart in
            the recorded_builds, those items are candidates to be reflected in Agile Central --> REFLECT
        """
        unrecorded_builds = []
        reflected_builds = []
        for ac_project, bld_data in bld_builds.items():
            for plan, builds in bld_data.items():
                for build in builds:
                    # look first for a matching project key in recorded_builds
                    if ac_project in recorded_builds:
                        plan_builds = recorded_builds[ac_project]
                        # now look for a matching job in plan_builds
                        plan_name = plan.name
                        if plan_name in plan_builds:
                            ac_build_nums = [int(bld.Number) for bld in plan_builds[plan_name]] #bld is a pyral build
                            if build.number in ac_build_nums:
                                reflected_builds.append((plan, build, ac_project))
                                continue
                    unrecorded_builds.append((plan, build, ac_project))
        return unrecorded_builds

    def getChangesetInfo(self, build):
        '''
            Endpoint:
                http://koljo03-s4576.ca.com:8085/rest/api/latest/result/FER-DON-251.json?expand=changes.change.files,vcsRevisions

            We did not find the following changeset data in json response:
              - uri of the changeset in VCS system
              - vcs type
            Mapping of vcs data we get from this endpoint to corresponding AC objects will require that we support
            additional configuration from the user, for example:
                       Projects:
                           - Project: Fernandel
                             AgileCentral_Project: Rally Fernandel
                             Plans:
                                 - Plan: DonCamillo
                                   Repositories:
                                       - Repository: don-camillo
                                         VCSType: bitbuket
                                       - Repository: koljo3
                                         VCSType: git


            The first release of Bamboo Build Connector will not support VCSData

        '''
        changesets = []

        bamboo_build_url = "%s.json?expand=changes.change.files,vcsRevisions" % build.link
        headers = {'Content-Type': 'application/json'}
        response, msg = None, None
        try:
            response = requests.get(bamboo_build_url, auth=self.creds, proxies=self.http_proxy, headers=headers)
        except Exception as msg:
            self.log.error(msg)
        if response.status_code >= 300:
            raise OperationalError('%s  status_code: %s' % (msg, response.status_code))

        result = response.json()
        changes = result['changes']['change']
        for change in changes:
            changesets.append(BambooChangeset(change))
        return changesets


####################################################################################################

class BambooPlan:
    def __init__(self, raw):
        self.full_name = raw['name']
        self.name      = raw['shortName']
        self.link      = raw['link']['href']
        self.key       = raw['key']
        self.url       = self.link.replace('rest/api/latest/plan', 'browse')
        self.project   = self.full_name.replace(' - %s' % self.name, '')
        self.path      = None

    def getActivityIdentifiers(self):
        return self.name, self.path, self.url

    def __str__(self):
        return "%s::%s" % (self.project, self.name)

###########################################################################################

class BambooBuild:
    def __init__(self, raw):
        """
        """
        self.id        = raw['id']
        self.number    = int(raw['number'])
        self.state     = raw['state']
        self.key       = raw['buildResultKey'] #FER-DON-45
        self.link      = raw['link']['href']  # http://koljo03-s4576.ca.com:8085/rest/api/latest/result/FER-DON-45
        self.url       = self.link.replace('rest/api/latest/result','browse')  # http://koljo03-s4576.ca.com:8085/browse/FER-DON-45
        self.finished  = raw['finished']
        self.plan      = BambooPlan(raw['plan'])
        self.started_time   = raw['buildStartedTime']
        self.completed_time = raw['buildCompletedTime']  # "2017-06-12T13:55:39.712-06:00
        self.started_timestamp   = TimeStamp.fromFormattedString(self.started_time,   ISO_FRAC_FORMAT)
        self.completed_timestamp = TimeStamp.fromFormattedString(self.completed_time, ISO_FRAC_FORMAT)
        self.project   = raw['projectName']
        self.duration  = int(raw['buildDuration'])
        self.changeset_behavior = 'match'
        self.changesets = []
        self.result     = self.state # all XBuild instances need a common result instance attribute

    def done(self):
        return self.finished

    def as_tuple_data(self):
        iso_str_start = self.started_timestamp.asISOString()
        matching_status = Matchmaker('Bamboo').matchStatus(str(self.state))
        build_data = [('Number', self.number),
                      ('Status', matching_status),
                      ('Start', iso_str_start),
                      ('Duration', self.duration / 1000.0),
                      ('Uri', self.url)]
        return build_data

    def __str__(self):
        duration = self.duration / 1000.0
        bstr = "%s Build # %5d   %-10.10s  Started: %s  Completed: %s   Duration: %s  URL: %s" % \
               (self.plan.name, self.number, self.result,
                self.started_timestamp.asLogStamp(), self.completed_timestamp.asLogStamp(),
                duration, self.url)
        return bstr


class BambooChangeset:
    def __init__(self, commit):
        self.commitId = commit['changesetId']
        commit_time   = commit['date']
        self.message  = commit['comment']
        self.timestamp = TimeStamp.fromFormattedString(commit_time, format=ISO_FRAC_FORMAT).asISOString()
        #self.vcs_type  = 'Git'
        #self.uri = commit['paths'][0]['file'] if commit['paths'] else '.'

    def __str__(self):
        cs_str = "   Commit ID # %s  Timestamp: %s  Message: %s " % \
                    (self.commitId, self.timestamp, self.message)
        return cs_str